#
# Guitar Synthesizer Example
#

from random import randint

# The default sampling rate used by JES is 22,050 Hz.
SAMPLE_RATE = 22050

# This function takes the frequency of a note as its input and returns a list
# that can be used as the cyclic queue for the Karplus-Strong string synthesis
# algorithm. The size of the list is computed as the sample size divided by
# the frequency of the note, and it is filled completely with random numbers
# to start. These random numbers represent the chaotic result of "plucking" a
# string.
#
def makeQueue(frequency):
  length = SAMPLE_RATE / frequency
  
  # This might be a good opportunity to show students how a list comprehension
  # is much more elegant than a loop to construct this list.
  queue = [randint(-32768, 32767) for i in range(length)]

  return queue


# This function performs one step of the Karplus-Strong algorithm on the list
# given as input, and then returns the sample value that was initially at the
# front of the list.
#
# In a nutshell, the algorithm removes the first value from the list, looks
# at (but does not remove) the second value (which is now the first), then
# averages these two values and multiplies them by a small decay factor.
# Finally, the result is appended back to the end of the list. This process
# simulates the settling of the string vibration.
#
def karplusStrongStep(queue):
  first = queue.pop(0)
  second = queue[0]
  result = int(0.996 * (first + second) / 2.0)
  queue.append(result)
  return first


# This function returns a sound that represents the guitar note at the
# given frequency (in Hz), played for the specified duration (in seconds).
#
def makeGuitarSound(frequency, duration):
  # Create the queue with random data for the frequency we want to make.
  queue = makeQueue(frequency)

  # Create an empty sound that will be used to store the synthesized samples.
  numSamples = int(duration * SAMPLE_RATE)
  sound = makeEmptySound(numSamples)
  
  # Repeatedly call the step method to simulate the vibration and get the next
  # sample in the sound and save it.
  for i in range(numSamples):
    nextSample = karplusStrongStep(queue)
    setSampleValueAt(sound, i, nextSample)

  return sound


# This is a naive first attempt at playing a song using the other functions
# that we've written. It takes as input a list of (frequency, duration)
# tuples and plays those notes in sequence. It's unsuitable for two major
# reasons:
#
# 1. Sounds are unceremoniously cut off after their duration has passed.
#    The result sounds harsh and artificial.
#
# 2. The processing required to generate a sound is slow (for a couple of
#    reasons -- Jython is slow, but we're also not using the list/queue in
#    the most efficient way). This creates a noticeable lag between notes.
#
# A good exercise for students would be to write a much better song-generating
# function. Rather than trying to generate each note on demand and play it,
# they should create one long sound that they will then copy the individual
# notes *into*. By keeping track of the starting point for each note, and even
# letting them overlap (by adding the samples together) if the length is
# shorter than a full decay, reasonably advanced music can be composed.
#
def badPlaySong(notes):
  for (frequency, duration) in notes:
    blockingPlay(makeGuitarSound(frequency, duration))


# A basic input loop to let us test the program.
prompt = 'Enter a note frequency in Hz (0 to stop):'
frequency = int(input(prompt))
while frequency != 0:
  sound = makeGuitarSound(frequency, 2)
  play(sound)
  frequency = int(input(prompt))

#badPlaySong([(440, 0.5), (220, 0.5), (570, 0.25), (345, 0.25), (700, 2)])
